from . import file_utilities
from . import display
from . import constants


'''
Try this in __init__.py
from constants import *
from display import *
from download import *
from file_utilities import *
from resources import *


And in modules try this, eg in display.py:
from . import file_utilities

'''